@@ -60,7 +60,7 @@ import Distribution.InstalledPackageInfo
          , emptyInstalledPackageInfo, parseInstalledPackageInfo )
 import Distribution.PackageDescription
         ( PackageDescription(..), BuildInfo(..), Library(..), Executable(..)
-        , hcOptions )
+        , hcOptions, usedExtensions )
 import Distribution.ModuleName (ModuleName)
 import qualified Distribution.ModuleName as ModuleName
 import Distribution.Simple.LocalBuildInfo
@@ -68,12 +68,13 @@ import Distribution.Simple.LocalBuildInfo
 import Distribution.Simple.BuildPaths
         ( mkLibName, objExtension, exeExtension )
 import Distribution.Simple.Compiler
-        ( CompilerFlavor(..), CompilerId(..), Compiler(..)
-        , Flag, extensionsToFlags, PackageDB(..), PackageDBStack )
+         ( CompilerFlavor(..), CompilerId(..), Compiler(..)
+         , Flag, languageToFlags, extensionsToFlags
+         , PackageDB(..), PackageDBStack )
 import qualified Distribution.Simple.PackageIndex as PackageIndex
 import Distribution.Simple.PackageIndex (PackageIndex)
 import Language.Haskell.Extension
-        ( Extension(..) )
+         ( Language(Haskell98), Extension(..) )
 import Distribution.Simple.Program
          ( ProgramConfiguration, userMaybeSpecifyPath, programPath
          , requireProgram, requireProgramVersion, lookupProgram
@@ -130,14 +131,17 @@ configure verbosity hcPath _hcPkgPath conf = do
 
   let comp = Compiler {
         compilerId         = CompilerId NHC nhcVersion,
+        compilerLanguages  = nhcLanguages,
         compilerExtensions = nhcLanguageExtensions
       }
   return (comp, conf'''')
 
+nhcLanguages :: [(Language, Flag)]
+nhcLanguages = [(Haskell98, "-98")]
+
 -- | The flags for the supported extensions
 nhcLanguageExtensions :: [(Extension, Flag)]
 nhcLanguageExtensions =
-    -- TODO: use -98 when no extensions are specified.
     -- NHC doesn't enforce the monomorphism restriction at all.
     -- TODO: pattern guards in 1.20
     [(NoMonomorphismRestriction, "")
@@ -273,7 +277,8 @@ buildLib verbosity pkg_descr lbi lib clbi = do
   let bi = libBuildInfo lib
       modules = exposedModules lib ++ otherModules bi
       -- Unsupported extensions have already been checked by configure
-      extensionFlags = extensionsToFlags (compiler lbi) (extensions bi)
+      languageFlags = languageToFlags (compiler lbi) (defaultLanguage bi)
+                   ++ extensionsToFlags (compiler lbi) (usedExtensions bi)
   inFiles <- getModulePaths lbi bi modules
   let targetDir = buildDir lbi
       srcDirs  = nub (map takeDirectory inFiles)
@@ -283,9 +288,9 @@ buildLib verbosity pkg_descr lbi lib clbi = do
        ["-hc=" ++ programPath nhcProg]
     ++ nhcVerbosityOptions verbosity
     ++ ["-d", targetDir, "-hidir", targetDir]
-    ++ extensionFlags
     ++ maybe [] (hcOptions NHC . libBuildInfo)
                            (library pkg_descr)
+    ++ languageFlags
     ++ concat [ ["-package", display (packageName pkgid) ]
               | (_, pkgid) <- componentPackageDeps clbi ]
     ++ inFiles
@@ -337,7 +342,8 @@ buildExe verbosity pkg_descr lbi exe clbi = do
   let bi = buildInfo exe
       modules = otherModules bi
       -- Unsupported extensions have already been checked by configure
-      extensionFlags = extensionsToFlags (compiler lbi) (extensions bi)
+      languageFlags = languageToFlags (compiler lbi) (defaultLanguage bi)
+                   ++ extensionsToFlags (compiler lbi) (usedExtensions bi)
   inFiles <- getModulePaths lbi bi modules
   let targetDir = buildDir lbi </> exeName exe
       exeDir    = targetDir </> (exeName exe ++ "-tmp")
@@ -348,9 +354,9 @@ buildExe verbosity pkg_descr lbi exe clbi = do
        ["-hc=" ++ programPath nhcProg]
     ++ nhcVerbosityOptions verbosity
     ++ ["-d", targetDir, "-hidir", targetDir]
-    ++ extensionFlags
     ++ maybe [] (hcOptions NHC . libBuildInfo)
                            (library pkg_descr)
+    ++ languageFlags
     ++ concat [ ["-package", display (packageName pkgid) ]
               | (_, pkgid) <- componentPackageDeps clbi ]
     ++ inFiles
