@@ -48,7 +48,6 @@ OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. -}
 
 module Distribution.Simple.Utils (
         cabalVersion,
-        cabalBootstrapping,
 
         -- * logging and errors
         die,
@@ -59,6 +58,7 @@ module Distribution.Simple.Utils (
 
         -- * running programs
         rawSystemExit,
+        rawSystemExitWithEnv,
         rawSystemStdout,
         rawSystemStdInOut,
         maybeExit,
@@ -179,6 +179,7 @@ import Distribution.Version
     (Version(..))
 
 import Control.Exception (evaluate)
+import System.Process (runProcess)
 
 #ifdef __GLASGOW_HASKELL__
 import Control.Concurrent (forkIO)
@@ -196,19 +197,18 @@ import Distribution.Compat.Exception
          ( catchIO, catchExit, onException )
 import Distribution.Verbosity
 
--- We only get our own version number when we're building with ourselves
-cabalVersion :: Version
-#ifdef CABAL_VERSION
-cabalVersion = Version [CABAL_VERSION] []
-#else
-cabalVersion = error "Cabal was not bootstrapped correctly"
+#ifdef VERSION_base
+import qualified Paths_Cabal (version)
 #endif
 
-cabalBootstrapping :: Bool
-#ifdef CABAL_VERSION
-cabalBootstrapping = False
+-- We only get our own version number when we're building with ourselves
+cabalVersion :: Version
+#if defined(VERSION_base)
+cabalVersion = Paths_Cabal.version
+#elif defined(CABAL_VERSION)
+cabalVersion = Version [CABAL_VERSION] []
 #else
-cabalBootstrapping = True
+cabalVersion = Version [1,9999] []  --used when bootstrapping
 #endif
 
 -- ----------------------------------------------------------------------------
@@ -346,6 +346,17 @@ printRawCommandAndArgs verbosity path args
  | verbosity >= verbose   = putStrLn $ unwords (path : args)
  | otherwise              = return ()
 
+printRawCommandAndArgsAndEnv :: Verbosity
+                             -> FilePath
+                             -> [String]
+                             -> [(String, String)]
+                             -> IO ()
+printRawCommandAndArgsAndEnv verbosity path args env
+ | verbosity >= deafening = do putStrLn ("Environment: " ++ show env)
+                               print (path, args)
+ | verbosity >= verbose   = putStrLn $ unwords (path : args)
+ | otherwise              = return ()
+
 -- Exit with the same exitcode if the subcommand fails
 rawSystemExit :: Verbosity -> FilePath -> [String] -> IO ()
 rawSystemExit verbosity path args = do
@@ -356,6 +367,20 @@ rawSystemExit verbosity path args = do
     debug verbosity $ path ++ " returned " ++ show exitcode
     exitWith exitcode
 
+rawSystemExitWithEnv :: Verbosity
+                     -> FilePath
+                     -> [String]
+                     -> [(String, String)]
+                     -> IO ()
+rawSystemExitWithEnv verbosity path args env = do
+    printRawCommandAndArgsAndEnv verbosity path args env
+    hFlush stdout
+    ph <- runProcess path args Nothing (Just env) Nothing Nothing Nothing
+    exitcode <- waitForProcess ph
+    unless (exitcode == ExitSuccess) $ do
+        debug verbosity $ path ++ " returned " ++ show exitcode
+        exitWith exitcode
+
 -- | Run a command and return its output.
 --
 -- The output is assumed to be text in the locale encoding.
