@@ -194,6 +194,7 @@ module Data.ByteString (
         hGetLine,               -- :: Handle -> IO ByteString
         hGetContents,           -- :: Handle -> IO ByteString
         hGet,                   -- :: Handle -> Int -> IO ByteString
+        hGetSome,               -- :: Handle -> Int -> IO ByteString
         hGetNonBlocking,        -- :: Handle -> Int -> IO ByteString
         hPut,                   -- :: Handle -> ByteString -> IO ()
         hPutStr,                -- :: Handle -> ByteString -> IO ()
@@ -250,12 +251,19 @@ import Data.Monoid              (Monoid, mempty, mappend, mconcat)
 import System.IO.Unsafe
 import qualified System.Environment
 import qualified System.IO      (hGetLine)
+import System.IO                (hIsEOF)
 #endif
 
 #if defined(__GLASGOW_HASKELL__)
 
 import System.IO                (hGetBufNonBlocking)
 
+#if MIN_VERSION_base(4,3,0)
+import System.IO                (hGetBufSome)
+#else
+import System.IO                (hWaitForInput, hIsEOF)
+#endif
+
 #if __GLASGOW_HASKELL__ >= 611
 import Data.IORef
 import GHC.IO.Handle.Internals
@@ -281,10 +289,17 @@ import GHC.ST                   (ST(..))
 
 -- An alternative to Control.Exception (assert) for nhc98
 #ifdef __NHC__
+
+import System.IO (Handle)
+
 #define assert  assertS "__FILE__ : __LINE__"
 assertS :: String -> Bool -> a -> a
 assertS _ True  = id
 assertS s False = error ("assertion failed at "++s)
+
+-- An alternative to hWaitForInput
+hWaitForInput :: Handle -> Int -> IO ()
+hWaitForInput _ _ = return ()
 #endif
 
 -- -----------------------------------------------------------------------------
@@ -1938,6 +1953,32 @@ hGetNonBlocking h i
 hGetNonBlocking = hGet
 #endif
 
+-- | Like 'hGet', except that a shorter 'ByteString' may be returned
+-- if there are not enough bytes immediately available to satisfy the
+-- whole request.  'hGetSome' only blocks if there is no data
+-- available, and EOF has not yet been reached.
+--
+hGetSome :: Handle -> Int -> IO ByteString
+hGetSome hh i
+#if MIN_VERSION_base(4,3,0)
+    | i >  0    = createAndTrim i $ \p -> hGetBufSome hh p i
+#else
+    | i >  0    = let
+                   loop = do
+                     s <- hGetNonBlocking hh i
+                     if not (null s)
+                        then return s
+                        else do eof <- hIsEOF hh
+                                if eof then return s
+                                       else hWaitForInput hh (-1) >> loop
+                                         -- for this to work correctly, the
+                                         -- Handle should be in binary mode
+                                         -- (see GHC ticket #3808)
+                  in loop
+#endif
+    | i == 0    = return empty
+    | otherwise = illegalBufferSize hh "hGetSome" i
+
 illegalBufferSize :: Handle -> String -> Int -> IO a
 illegalBufferSize handle fn sz =
     ioError (mkIOError illegalOperationErrorType msg (Just handle) Nothing)
