@@ -9,7 +9,7 @@
 #ifndef PROFHEAP_H
 #define PROFHEAP_H
 
-BEGIN_RTS_PRIVATE
+#include "BeginPrivate.h"
 
 void    heapCensus         (void);
 nat     initHeapProfiling  (void);
@@ -17,6 +17,6 @@ void    endHeapProfiling   (void);
 void    LDV_recordDead     (StgClosure *c, nat size);
 rtsBool strMatchesSelector (char* str, char* sel);
 
-END_RTS_PRIVATE
+#include "EndPrivate.h"
 
 #endif /* PROFHEAP_H */
