@@ -467,10 +467,8 @@ genApply regstatus args =
 --        print "    [THUNK_STATIC]    &&thunk_lbl,"
 --        print "    [THUNK_SELECTOR]  &&thunk_lbl,"
 --        print "    [IND]	      &&ind_lbl,"
---        print "    [IND_OLDGEN]      &&ind_lbl,"
 --        print "    [IND_STATIC]      &&ind_lbl,"
 --        print "    [IND_PERM]	      &&ind_lbl,"
---        print "    [IND_OLDGEN_PERM] &&ind_lbl"
 --        print "  };"
     
        tickForArity (length args),
@@ -564,8 +562,8 @@ genApply regstatus args =
 --    else:
 	text "case AP,",
 	text "     AP_STACK,",
-	text "     CAF_BLACKHOLE,",
 	text "     BLACKHOLE,",
+	text "     WHITEHOLE,",
         text "     THUNK,",
         text "     THUNK_1_0,",
         text "     THUNK_0_1,",
@@ -590,10 +588,8 @@ genApply regstatus args =
 --        print "    ind_lbl:"
 --    else:
         text "case IND,",
-        text "     IND_OLDGEN,",
         text "     IND_STATIC,",
-        text "     IND_PERM,",
-        text "     IND_OLDGEN_PERM: {",
+        text "     IND_PERM: {",
 	nest 4 (vcat [
 	  text "R1 = StgInd_indirectee(R1);",
             -- An indirection node might contain a tagged pointer
